# -*- coding: utf-8 -*-


class RemoteClient(object):

    def __init__(self):
        pass

    def connect(self, endpoint):
        pass

    def send(self, message):
        pass

    def recv(self, timeout=None):
        pass




# -*- coding: utf-8 -*-
import json


class MessageHandler(object):
    """
    A message is basically a request (REQ) from a client, a handler is a
    function that resolves the request and returns a reply (REP).
    This class maps a message with a callable function (the handler).

    The handler evaluation order is important and it's determined by the
    following rules (implemented in the __lt__ method):

    1. More properties win.
        {a:1, b:2} wins over {a:1} as it has more properties.

    2. If same number of properties, alphabetical order win.
        {a:1, b:2} wins over {a:1, c:3} as b comes before c alphabetically.
        {a:1, b:2, d:4} wins over {a:1, c:3, d:4} as b comes before c.

    3. If same properties, local handlers win.
        The local handler {a:1, b:2} wins over the remote one {a:1, b:2}.
    """
    def __init__(self, message, handler):
        self.message = message
        self.handler = handler
        self.key = json.dumps(message, sort_keys=True)

    @property
    def is_local(self):
        return True

    def __lt__(self, other):
        """
        This function implements the precedence order of message handlers:
          1. More properties win.
          2. If same number of properties, alphabetical order win.
          3. If same properties, local handlers win.
        """
        if len(self.message) > len(other.message):
            return True
        if len(self.message) == len(other.message) and self.key < other.key:
            return True
        if self.message == other.message and self.is_local:
            return True
        return False

    def __str__(self):
        return self.key

    def match_in(self, message):
        try:
            for key, value in self.message.items():
                assert (message[key] == value)
            return True
        except (AssertionError, KeyError):
            return False



# -*- coding: utf-8 -*-
import collections
import re
import threading
import zmq

from lucena.exceptions import WorkerAlreadyStarted, WorkerNotStarted, \
    LookupHandlerError
from lucena.io2.socket import Socket
from lucena.message_handler import MessageHandler


class Worker(object):

    RunningWorker = collections.namedtuple(
        'RunningWorker',
        ['worker', 'thread']
    )

    PollHandler = collections.namedtuple(
        'PollHandler',
        ['socket', 'flags', 'handler']
    )

    class Controller(object):

        def __init__(self, *args, **kwargs):
            self.context = zmq.Context.instance()
            self.args = args
            self.kwargs = kwargs
            self.poller = zmq.Poller()
            self.running_workers = None
            self.control_socket = Socket(self.context, zmq.ROUTER)
            self.control_socket.bind(Socket.inproc_unique_endpoint())

        def is_started(self):
            return self.running_workers is not None

        def start(self, number_of_workers=1):
            if self.is_started():
                raise WorkerAlreadyStarted()
            if not isinstance(number_of_workers, int) or number_of_workers < 1:
                raise ValueError("Parameter number_of_workers must be a positive integer.")
            self.running_workers = {}
            for i in range(number_of_workers):
                worker = Worker(*self.args, **self.kwargs)
                thread = threading.Thread(
                    target=worker,
                    daemon=False,
                    kwargs={
                        'endpoint': self.control_socket.last_endpoint,
                        'index': i
                    }
                )
                thread.start()
                identity, client, message = self.recv()
                assert identity == worker.identity(i)
                assert client == b'$controller'
                assert message == {"$signal": "ready"}
                self.running_workers[identity] = Worker.RunningWorker(worker, thread)
            return list(self.running_workers.keys())

        def stop(self, timeout=None):
            for worker_id, running_worker in self.running_workers.items():
                self.send(worker_id, b'$controller', {'$signal': 'stop'})
                _worker_id, client, message = self.recv()
                assert _worker_id == worker_id
                assert client == b'$controller'
                assert message == {'$signal': 'stop', '$rep': 'OK'}
                running_worker.thread.join(timeout=timeout)
            self.running_workers = None

        def send(self, worker_id, client_id, message):
            if not self.is_started():
                raise WorkerNotStarted()
            return self.control_socket.send_to_worker(worker_id, client_id, message)

        def recv(self):
            # TODO: Add timeout
            if not self.is_started():
                raise WorkerNotStarted()
            worker, client, message = self.control_socket.recv_from_worker()
            return worker, client, message

    # Worker implementation.

    def __init__(self, *args, **kwargs):
        self.context = zmq.Context.instance()
        self.poller = zmq.Poller()
        self.poll_handlers = []
        self.message_handlers = []
        self.bind_handler({}, self.handler_default)
        self.bind_handler({'$signal': 'stop'}, self.handler_stop)
        self.bind_handler({'$req': 'eval'}, self.handler_eval)
        self.stop_signal = False
        self.control_socket = None

    def __call__(self, endpoint, index):
        self._before_start(self.identity(index))
        self.control_socket.connect(endpoint)
        self.control_socket.send_to_client(b'$controller', {"$signal": "ready"})
        while not self.stop_signal:
            self._handle_poll()
        self._before_stop()

    def _add_poll_handler(self, socket, flags, handler):
        poll_handler = self.PollHandler(socket, flags, handler)
        self.poll_handlers.append(poll_handler)

    def _before_start(self, identity):
        self.poll_handlers = []
        self.stop_signal = False
        self.control_socket = Socket(self.context, zmq.REQ, identity=identity)
        self._add_poll_handler(
            self.control_socket,
            zmq.POLLIN if not self.stop_signal else 0,
            self._handle_ctrl_socket
        )

    def _before_stop(self):
        self.control_socket.close()

    def _handle_poll(self):
        for poll_handler in self.poll_handlers:
            self.poller.register(
                poll_handler.socket,
                poll_handler.flags
            )
        sockets = dict(self.poller.poll(1))
        for poll_handler in self.poll_handlers:
            if poll_handler.socket in sockets:
                poll_handler.handler()

    def _handle_ctrl_socket(self):
        client, message = self.control_socket.recv_from_client()
        response = self.resolve(message)
        self.control_socket.send_to_client(client, response)

    @classmethod
    def identity(cls, index=0):
        id1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', cls.__name__)
        id2 = re.sub('([a-z0-9])([A-Z])', r'\1_\2', id1).lower()
        return '{}#{}'.format(id2, index).encode('utf8')

    @staticmethod
    def handler_default(message):
        response = {}
        response.update(message)
        response.update({"$rep": None, "$error": "No handler match"})
        return response

    def handler_eval(self, message):
        response = {}
        response.update(message)
        attr = getattr(self, message.get('$attr'))
        response.update({'$rep': attr})
        return response

    def handler_stop(self, message):
        response = {}
        response.update(message)
        response.update({'$rep': 'OK'})
        self.stop_signal = True
        return response

    def bind_handler(self, message, handler):
        self.message_handlers.append(MessageHandler(message, handler))
        self.message_handlers.sort()

    def unbind_handler(self, message):
        for message_handler in self.message_handlers:
            if message_handler.message == message:
                self.message_handlers.remove(message_handler)
                return
        raise LookupHandlerError("No handler for {}".format(message))

    def get_handler_for(self, message):
        for message_handler in self.message_handlers:
            if message_handler.match_in(message):
                return message_handler.handler
        raise LookupHandlerError("No handler for {}".format(message))

    def resolve(self, message):
        handler = self.get_handler_for(message)
        return handler(message)




# -*- coding: utf-8 -*-
import tempfile
import threading

import zmq

from lucena.exceptions import ServiceAlreadyStarted, ServiceNotStarted
from lucena.io2.socket import Socket
from lucena.worker import Worker


class Service(Worker):

    class Controller(object):

        def __init__(self, *args, **kwargs):
            self.context = zmq.Context.instance()
            self.args = args
            self.kwargs = kwargs
            self.poller = zmq.Poller()
            self.service_identity = Service.identity()
            self.service_thread = None
            self.control_socket = Socket(self.context, zmq.ROUTER)
            self.control_socket.bind(Socket.inproc_unique_endpoint())

        def is_started(self):
            return self.service_thread is not None

        def start(self):
            if self.service_thread is not None:
                raise ServiceAlreadyStarted()
            service = Service(*self.args, **self.kwargs)
            self.service_thread = threading.Thread(
                target=service,
                daemon=False,
                kwargs={
                    'endpoint': self.control_socket.last_endpoint,
                    'index': 0
                }
            )
            self.service_thread.start()
            message = self.recv()
            assert message == {"$signal": "ready"}

        def stop(self, timeout=None):
            self.send({'$signal': 'stop'})
            message = self.recv()
            assert message == {'$signal': 'stop', '$rep': 'OK'}
            self.service_thread.join(timeout=timeout)
            self.service_thread = None

        def send(self, message):
            if not self.is_started():
                raise ServiceNotStarted()
            return self.control_socket.send_to_worker(
                self.service_identity,
                b'$controller',
                message
            )

        def recv(self):
            # TODO: Add timeout
            if not self.is_started():
                raise ServiceNotStarted()
            worker, client, message = self.control_socket.recv_from_worker()
            assert client == b'$controller'
            assert worker == self.service_identity
            return message

    # Service implementation.

    def __init__(self, worker_factory, endpoint=None, number_of_workers=1):
        # http://zguide.zeromq.org/page:all#Getting-the-Context-Right
        # You should create and use exactly one context in your process.
        super(Service, self).__init__()
        self.worker_factory = worker_factory
        self.endpoint = endpoint if endpoint is not None \
            else "ipc://{}.ipc".format(tempfile.NamedTemporaryFile().name)
        self.number_of_workers = number_of_workers
        self.socket = None
        self.worker_controller = None
        self.worker_ready_ids = None
        self.total_client_requests = 0

    def _before_start(self, identity):
        super(Service, self)._before_start(identity)
        self.worker_ready_ids = []
        self.socket = Socket(self.context, zmq.ROUTER)
        self.socket.bind(self.endpoint)
        self.worker_controller = Worker.Controller(self.worker_factory)
        self.worker_ready_ids = self.worker_controller.start(self.number_of_workers)
        self._add_poll_handler(
            self.socket,
            zmq.POLLIN if self.worker_ready_ids else 0,
            self._handle_socket
        )
        self._add_poll_handler(
            self.worker_controller.control_socket,
            zmq.POLLIN,
            self._handle_worker_controller
        )

    def _before_stop(self):
        super(Service, self)._before_stop()
        self.socket.close()
        self.worker_controller.stop()

    def _handle_socket(self):
        assert len(self.worker_ready_ids) > 0
        client, request = self.socket.recv_from_client()
        worker_name = self.worker_ready_ids.pop(0)
        self.worker_controller.send(worker_name, client, request)
        self.total_client_requests += 1

    def _handle_worker_controller(self):
        worker_id, client, reply = self.worker_controller.recv()
        self.worker_ready_ids.append(worker_id)
        self.socket.send_to_client(client, reply)

    def pending_workers(self):
        return self.worker_ready_ids is not None and \
               len(self.worker_ready_ids) < self.number_of_workers


def create_service(worker_factory=None, endpoint=None, number_of_workers=1):
    return Service.Controller(
        worker_factory,
        endpoint=endpoint,
        number_of_workers=number_of_workers
    )



# -*- coding: utf-8 -*-


class LucenaException(Exception):
    """Lucena has raised an error."""

    def __str__(self):
        if self.args or not self.__doc__:
            return super(LucenaException, self).__str__()
        return self.__doc__


class LookupHandlerError(LucenaException):
    """Unable to resolve this message."""
    pass


class WorkerAlreadyStarted(LucenaException):
    """This Worker has already been started."""
    pass


class WorkerNotStarted(LucenaException):
    """This Worker has not been started."""
    pass


class ServiceAlreadyStarted(LucenaException):
    """This Service has already been started."""
    pass


class ServiceNotStarted(LucenaException):
    """This Service has not been started."""
    pass
